{% extends "sphinxdoc/layout.html" %}

{% block rootrellink %}
    <li><a href="{{ pathto('index') }}">home</a>|&nbsp;</li>
    <li><a href="{{ pathto('users/faq') }}">FAQ</a>|&nbsp;</li>
    <li><a href="{{ pathto('devel/examples/examples') }}">examples</a>|&nbsp;</li>
    <li><a href="{{ pathto('users/screenshots') }}">screenshots</a>|&nbsp;</li>
    <li><a href="{{ pathto('contents') }}">documentation </a> &raquo;</li>
{% endblock %}

{% block relbar1 %}
{{ super() }}
{% endblock %}


{% block sidebar2 %}
{{ super() }}
{% endblock %}
